from collections import namedtuple

import onnx


Graph = namedtuple('Graph', ['generate_name','inputs', 'outputs', 'transients', 'nodes', 'initializers', 'name'])
Graph.__new__.__defaults__ = ([], [], [], [], [], "")


def create_name_generator():
    state = {}
    def _generate_unique_name(name):
        if name in state:
            i = state[name]
            state[name] += 1
        else:
            state[name] = 0
            
        return "{}_{}".format(name, state[name])
        
        
    return _generate_unique_name


def extend(i, val):
    """Extends a list as a copy
    """
    ii = list(i)
    ii.extend(val)
    return ii


def pipe(*args):
    pass


def create_graph(name=""):
    return Graph(
        name=name,
        generate_name=create_name_generator()
    )

def compile(graph):
    #outputs = graph.transients

    graph = onnx.helper.make_graph(
        nodes=graph.nodes,
        name=graph.name,
        inputs=graph.inputs,    
        outputs=graph.outputs,
        initializer=graph.initializers,
    )
    return onnx.helper.make_model(graph, producer_name='ebm2onnx')

def create_input(graph, name, type, shape):
    input = onnx.helper.make_tensor_value_info(name , type, shape)
    return Graph(
        generate_name=graph.generate_name,
        inputs=[input],
        transients=[input],
    )


def add_output(graph, name, type, shape):
    output = onnx.helper.make_tensor_value_info(name , type, shape)
    return graph._replace(
        outputs=extend(graph.outputs, [output]),
    )


def create_initializer(graph, name, type, shape, value):
    initializer = onnx.helper.make_tensor(graph.generate_name(name) , type, shape, value)
    return Graph(
        generate_name=graph.generate_name,
        initializers=[initializer],
        transients=[initializer],
    )


def strip_to_transients(graph):
    return Graph(
        generate_name=graph.generate_name,
        transients=graph.transients,
    )


def merge(*args):
    g = None

    for graph in args:
        if g is None:
            g = graph
        else:
            g = g._replace(
                inputs=extend(g.inputs, graph.inputs),
                outputs=extend(g.outputs, graph.outputs),
                initializers=extend(g.initializers, graph.initializers),
                transients=extend(g.transients, graph.transients),
                nodes=extend(g.nodes, graph.nodes),
            )

    return g


import onnx
import ebm2onnx.graph as graph


def category_mapper(cats_int64s, cats_strings, default_int64=-1, default_string="_Unused"):
    def _category_mapper(g):
        category_mapper_result_name = g.generate_name('category_mapper_result')
        nodes = [
            onnx.helper.make_node(
                "CategoryMapper",
                [g.transients[0].name], [category_mapper_result_name],
                cats_int64s=cats_int64s,
                cats_strings=cats_strings,
                default_int64=default_int64,
                default_string=default_string,
                domain='ai.onnx.ml',
            ),
        ]

        return g._replace(
            nodes=graph.extend(g.nodes, nodes),
            transients=[
                 onnx.helper.make_tensor_value_info(category_mapper_result_name, onnx.TensorProto.UNDEFINED, []),
            ],
        )

    return _category_mapper



import numpy as np
import onnx
import ebm2onnx.operators as ops
import ebm2onnx.operators_ml as mlops
import ebm2onnx.graph as graph


def get_bin_index_on_continuous_value(bin_edges):
    """

    Input graph
    """
    def _get_bin_index_on_continuous_value(g):
        bin_count = len(bin_edges)
        index_range = list(range(bin_count))
        
        init_bin_index_range = graph.create_initializer(g, "bin_index_range", onnx.TensorProto.FLOAT, [bin_count], index_range)
        init_bin_edges = graph.create_initializer(g, "bin_edges", onnx.TensorProto.DOUBLE, [bin_count], bin_edges)

        g = ops.cast(onnx.TensorProto.DOUBLE)(g)
        g = ops.less_or_equal()(graph.merge(init_bin_edges, g))
        g = ops.cast(onnx.TensorProto.FLOAT)(g)
        g = ops.mul()(graph.merge(g, init_bin_index_range))
        g = ops.argmax(axis=1)(g)
        return g

    return _get_bin_index_on_continuous_value


def get_bin_index_on_categorical_value(col_mapping, missing_str=str(np.nan)):
    def _get_bin_index_on_categorical_value(g):
        ints = [0]
        strings = [missing_str]
        for k, v in col_mapping.items():
            ints.append(v)
            strings.append(k)

        g = mlops.category_mapper(
            cats_int64s=ints,
            cats_strings=strings,
        )(g)
        g = ops.flatten()(g)
        return g

    return _get_bin_index_on_categorical_value


def get_bin_score_1d(bin_scores):
    if len(bin_scores.shape) == 1:
        bin_scores= bin_scores.reshape((-1, 1))

    def _get_bin_score_1d(g):
        init_bin_scores = graph.create_initializer(
            g, "bin_scores", onnx.TensorProto.FLOAT,
            bin_scores.shape, bin_scores.flatten())

        init_reshape = graph.create_initializer(
            g, "score_reshape", onnx.TensorProto.INT64,
            [3], [-1, 1, bin_scores.shape[1]],
        )

        g = ops.gather_nd()(graph.merge(init_bin_scores, g))  # gather score for each class
        g = ops.reshape()(graph.merge(g, init_reshape))
        return g

    return _get_bin_score_1d


def get_bin_score_2d(bin_scores):
    def _get_bin_score_2d(g):
        init_bin_scores = graph.create_initializer(
            g, "bin_scores", onnx.TensorProto.FLOAT,
            [bin_scores.shape[0], bin_scores.shape[1]],
            bin_scores.flatten(),
        )

        init_reshape = graph.create_initializer(
            g, "score_reshape", onnx.TensorProto.INT64,
            [3], [-1, 1, 1],
        )

        g = ops.concat(axis=1)(g)
        g = ops.gather_nd()(graph.merge(init_bin_scores, g))
        g = ops.reshape()(graph.merge(g, init_reshape))
        return g

    return _get_bin_score_2d


def compute_class_score(intercept):
    """
    intercept shape: [class_count]
    input shapes: [batch_size x 1 x class_count]
    output shape: [batch_size x class_count]

    score output shape: [batch_size x feature_count x class_count]
    """
    def _compute_class_score(g):
        init_intercept = graph.create_initializer(
            g, "intercept", onnx.TensorProto.FLOAT,
            [intercept.shape[0]], intercept,
        )
        init_sum_axis = graph.create_initializer(
            g, "sum_axis", onnx.TensorProto.INT64,
            [1], [1],
        )

        g = ops.concat(axis=1)(g)
        scores_output_name = g.transients[0].name
        g = ops.reduce_sum(keepdims=0)(graph.merge(g, init_sum_axis))
        g = ops.add()(graph.merge(g, init_intercept))
        return g, scores_output_name

    return _compute_class_score


def predict_class(binary):
    def _predict_class(g):
        if binary is True:
            init_zeros = graph.create_initializer(
                g, "zeros", onnx.TensorProto.FLOAT,
                [2], [0.0, 1.0],
            )

            g = ops.mul()(graph.merge(g, init_zeros))

        init_reshape = graph.create_initializer(
            g, "reshape", onnx.TensorProto.INT64,
            [1], [0],
        )

        g = ops.argmax(axis=1)(g)
        g = ops.reshape()(graph.merge(g, init_reshape))
        g = ops.identity("predict")(g)
        return g

    return _predict_class


def predict_proba(binary):
    def _predict_proba(g):
        if binary is True:
            init_zeros = graph.create_initializer(
                g, "zeros", onnx.TensorProto.FLOAT,
                [2], [0.0, 1.0],
            )

            init_reshape = graph.create_initializer(
                g, "reshape", onnx.TensorProto.INT64,
                [1], [0],
            )

            g = ops.mul()(graph.merge(g, init_zeros))
        g = ops.softmax(axis=1)(g)
        g = ops.identity("predict")(g)
        return g

    return _predict_proba


def predict_value():
    """Final prediction step for regression

    No operations are needed here, we just reshape the
    [None, 1] scores to [None].
    """
    def _predict_value(g):
        init_reshape = graph.create_initializer(
            g, "reshape", onnx.TensorProto.INT64,
            [1], [0],
        )

        g = ops.reshape()(graph.merge(g, init_reshape))
        g = ops.identity("predict")(g)
        return g

    return _predict_value



from collections import namedtuple
from .utils import get_latest_opset_version
from ebm2onnx import graph
from ebm2onnx import ebm
import ebm2onnx.operators as ops

import numpy as np
import onnx

from interpret.glassbox import ExplainableBoostingClassifier, ExplainableBoostingRegressor


onnx_type_for={
    'float': onnx.TensorProto.FLOAT,
    'double': onnx.TensorProto.DOUBLE,
    'int': onnx.TensorProto.INT64,
    'str': onnx.TensorProto.STRING,
}


def infer_features_dtype(dtype, feature_name):
    feature_dtype = onnx.TensorProto.DOUBLE
    if dtype is not None:
        feature_dtype = onnx_type_for[dtype[feature_name]]

    return feature_dtype

def to_onnx(model, dtype, name=None,
            predict_proba=False,
            explain=False
            ):
    """Converts an EBM model to ONNX

    Args:
        model: The EBM model, trained with interpretml
        dtype: A dict containing the type of each input feature. Types are expressed as strings, the following values are supported: float, double, int, str.
        name: [Optional] The name of the model
        predict_proba: [Optional] For classification models, output prediction probabilities instead of class
        explain: [Optional] Adds an additional output with the score per feature per class

    Returns:
        An ONNX model.
    """
    #target_opset = target_opset or get_latest_opset_version()
    root = graph.create_graph(name=name)

    class_index=0
    inputs = [None for _ in model.feature_names]
    parts = []

    # first compute the score of each feature
    for feature_index in range(len(model.feature_names)):
        feature_name=model.feature_names[feature_index]
        feature_type=model.feature_types[feature_index]
        feature_group=model.feature_groups_[feature_index]

        if feature_type == 'continuous':
            bins = [np.NINF, np.NINF] + list(model.preprocessor_.col_bin_edges_[feature_group[0]])
            additive_terms = model.additive_terms_[feature_index]

            feature_dtype = infer_features_dtype(dtype, feature_name)
            part = graph.create_input(root, feature_name, feature_dtype, [None])
            part = ops.flatten()(part)
            inputs[feature_index] = part
            part = ebm.get_bin_index_on_continuous_value(bins)(part)
            part = ebm.get_bin_score_1d(additive_terms)(part)
            parts.append(part)

        elif feature_type == 'categorical':
            col_mapping = model.preprocessor_.col_mapping_[feature_group[0]]
            additive_terms = model.additive_terms_[feature_index]

            part = graph.create_input(root, feature_name, onnx.TensorProto.STRING, [None])
            part = ops.flatten()(part)
            inputs[feature_index] = part
            part = ebm.get_bin_index_on_categorical_value(col_mapping)(part)
            part = ebm.get_bin_score_1d(additive_terms)(part)
            parts.append(part)

        elif feature_type == 'interaction':
            i_parts = []
            for index in range(2):
                i_feature_index = feature_group[index]
                i_feature_type = model.feature_types[i_feature_index]

                if i_feature_type == 'continuous':
                    bins = [np.NINF, np.NINF] + list(model.pair_preprocessor_.col_bin_edges_[i_feature_index])
                    input = graph.strip_to_transients(inputs[i_feature_index])
                    i_parts.append(ebm.get_bin_index_on_continuous_value(bins)(input))

                elif i_feature_type == 'categorical':
                    col_mapping = model.preprocessor_.col_mapping_[i_feature_index]
                    input = graph.strip_to_transients(inputs[i_feature_index])
                    i_parts.append(ebm.get_bin_index_on_categorical_value(col_mapping)(input))

                else:
                    raise NotImplementedError(f"feature type {feature_type} is not supported in interactions")

            part = graph.merge(*i_parts)
            additive_terms = model.additive_terms_[feature_index]
            part = ebm.get_bin_score_2d(np.array(additive_terms))(part)
            parts.append(part)

        else:
            raise NotImplementedError(f"feature type {feature_type} is not supported")

    # compute scores, predict and proba
    g = graph.merge(*parts)
    if type(model) is ExplainableBoostingClassifier:
        g, scores_output_name = ebm.compute_class_score(model.intercept_)(g)
        if len(model.classes_) == 2: # binary classification
            if predict_proba is False:
                g = ebm.predict_class(binary=True)(g)
                g = graph.add_output(g, g.transients[0].name, onnx.TensorProto.INT64, [None])
            else:
                g = ebm.predict_proba(binary=True)(g)
                g = graph.add_output(g, g.transients[0].name, onnx.TensorProto.FLOAT, [None, len(model.classes_)])
        else:
            if predict_proba is False:
                g = ebm.predict_class(binary=False)(g)
                g = graph.add_output(g, g.transients[0].name, onnx.TensorProto.INT64, [None])
            else:
                g = ebm.predict_proba(binary=False)(g)
                g = graph.add_output(g, g.transients[0].name, onnx.TensorProto.FLOAT, [None, len(model.classes_)])
    elif type(model) is ExplainableBoostingRegressor:
        g, scores_output_name = ebm.compute_class_score(np.array([model.intercept_]))(g)
        g = ebm.predict_value()(g)
        g = graph.add_output(g, g.transients[0].name, onnx.TensorProto.FLOAT, [None])
    else:
        raise NotImplementedError("{} models are not supported".format(type(model)))

    if explain is True:
        if len(model.classes_) == 2:
            g = graph.add_output(g, scores_output_name, onnx.TensorProto.FLOAT, [None, len(model.feature_names)])
        else:
            g = graph.add_output(g, scores_output_name, onnx.TensorProto.FLOAT, [None, len(model.feature_names), len(model.classes_)])

    model = graph.compile(g)
    return model



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'onnx>=1.8',
    'interpret-core[required,ebm]>=0.2',
 ]

setup(
    author="Romain Picard",
    author_email='romain.picard@softathome.com',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="EBM model serialization to ONNX",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='ebm2onnx',
    name='ebm2onnx',
    packages=find_packages(include=['ebm2onnx']),
    url='https://github.com/SoftAtHome/ebm2onnx.git',
    version='1.0.0',
    zip_safe=True,
)
